import { RequestEvent, Handle } from '@sveltejs/kit';
import { createClient } from 'redis';
import { Firestore } from 'firebase-admin/firestore';

interface KeyInfoData {
    user: string;
    name: string;
    description: string;
    expires: Date | null;
    permissions: string[];
}
interface KeyInfo extends KeyInfoData {
    hash: string;
}

interface KeyStore {
    put(hash: string, info: KeyInfoData): Promise<void>;
    get(hash: string): Promise<KeyInfoData | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<KeyInfo[]>;
}

declare const SECOND = 1;
declare const MINUTE: number;
declare const HOUR: number;
declare const DAY: number;
declare const WEEK: number;
interface Refill {
    readonly rate: number;
    readonly size: number;
}
declare function parseRefill(value: string): {
    rate: number;
    size: number;
};

interface BucketResult {
    allowed: boolean;
    limit: number;
    remaining: number;
    reset: number;
    policy: string;
}
interface TokenBuckets {
    consume(key: string, refill: Refill, count?: number): Promise<BucketResult>;
}
declare abstract class TokenBucket {
    protected abstract consume(key: string, refill: Refill, count?: number): Promise<BucketResult>;
    protected apply(refill: Refill, count: number, allowed: boolean, tokens: number): {
        allowed: boolean;
        limit: number;
        remaining: number;
        reset: number;
        policy: string;
    };
}

type Input = Parameters<Handle>[0];
type CustomFn = (event: RequestEvent, key: string | null) => Promise<string | null>;
interface BaseParam {
    searchParam?: string;
    httpHeader?: string;
    cookie?: string;
    custom?: CustomFn;
    key_length?: number;
}
interface SearchParam extends BaseParam {
    searchParam: string;
}
interface HttpHeader extends BaseParam {
    httpHeader: string;
}
interface Cookie extends BaseParam {
    cookie: string;
}
interface Custom extends BaseParam {
    custom: CustomFn;
}
type Options = SearchParam | HttpHeader | Cookie | Custom;
declare class ApiKeys {
    private readonly storage;
    private readonly buckets;
    private readonly options;
    constructor(storage: KeyStore, buckets: TokenBuckets, options?: Options);
    private hashFromBytes;
    generate(info: KeyInfoData): Promise<{
        key: string;
        hash: string;
    }>;
    validate(key: string | null): Promise<{
        user: string;
        name: string;
        description: string;
        expires: Date | null;
        permissions: string[];
        hash: string;
    } | null>;
    remove(hash: string): Promise<void>;
    retrieve(user: string): Promise<KeyInfo[]>;
    extract(event: RequestEvent): Promise<string | null>;
    handle: (input: Input) => Promise<Response>;
}

declare class InMemoryTokenBucket extends TokenBucket {
    protected readonly buckets: Map<string, {
        tokens: number;
        updated: number;
    }>;
    consume(key: string, refill: Refill, count?: number): Promise<{
        allowed: boolean;
        limit: number;
        remaining: number;
        reset: number;
        policy: string;
    }>;
}

type Redis$1 = ReturnType<typeof createClient>;
declare class RedisTokenBucket extends TokenBucket {
    private readonly redis;
    private readonly prefix;
    private readonly module;
    private readonly fname;
    private constructor();
    consume(key: string, refill: Refill, count?: number): Promise<{
        allowed: boolean;
        limit: number;
        remaining: number;
        reset: number;
        policy: string;
    }>;
    static create(redis: Redis$1, prefix?: string, module?: string, fname?: string): Promise<RedisTokenBucket>;
}

declare class LruCacheKeyStore implements KeyStore {
    private readonly store;
    private cache;
    constructor(store: KeyStore, maxSize?: number, // default capacity
    ttl?: number);
    put(hash: string, info: KeyInfoData): Promise<void>;
    get(hash: string): Promise<KeyInfoData | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<KeyInfo[]>;
}

declare class FirestoreKeyStore implements KeyStore {
    private readonly firestore;
    private readonly name;
    private readonly collection;
    constructor(firestore: Firestore, name?: string);
    put(hash: string, info: KeyInfoData): Promise<void>;
    get(hash: string): Promise<KeyInfo | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<KeyInfo[]>;
}

declare class InMemoryKeyStore implements KeyStore {
    private keys;
    put(hash: string, info: KeyInfoData): Promise<void>;
    get(hash: string): Promise<KeyInfoData | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<{
        user: string;
        name: string;
        description: string;
        expires: Date | null;
        permissions: string[];
        hash: string;
    }[]>;
}

type Redis = ReturnType<typeof createClient>;
declare class RedisKeyStore implements KeyStore {
    private readonly redis;
    private readonly prefix;
    static create(redis: Redis, prefix?: string): Promise<RedisKeyStore>;
    private constructor();
    put(hash: string, info: KeyInfoData): Promise<void>;
    get(hash: string): Promise<KeyInfoData | null>;
    del(hash: string): Promise<void>;
    list(user: string): Promise<{
        hash: string;
        user: string;
        name: string;
        description: string;
        permissions: string[];
        expires: Date | null;
    }[]>;
}

declare class Api {
    private readonly event;
    private readonly bucket;
    readonly key: string | null;
    readonly info: KeyInfo | null;
    private _name;
    private _cost;
    private _anon;
    private _has;
    private _any;
    private _all;
    readonly SECOND = 1;
    readonly MINUTE: number;
    readonly HOUR: number;
    readonly DAY: number;
    readonly WEEK: number;
    constructor(event: RequestEvent, bucket: TokenBuckets, key: string | null, info: KeyInfo | null);
    anonymous(): this;
    name(name: string): this;
    cost(cost: number): this;
    has(permission: string): this;
    any(permissions: string[]): this;
    all(permissions: string[]): this;
    limit(refill: Refill): Promise<BucketResult>;
}

interface ApiLocals {
    api: Api;
}
declare global {
    namespace App {
        interface Locals extends ApiLocals {
        }
    }
}

export { ApiKeys, type ApiLocals, type BucketResult, type Cookie, type Custom, DAY, FirestoreKeyStore, HOUR, type HttpHeader, InMemoryKeyStore, InMemoryTokenBucket, type KeyInfo, type KeyInfoData, type KeyStore, LruCacheKeyStore, MINUTE, type Options, RedisKeyStore, RedisTokenBucket, type Refill, SECOND, type SearchParam, TokenBucket, type TokenBuckets, WEEK, parseRefill };
